/*
 * @(#)JControlSet.java		1.10 07/01/20
 *
 * ChargedFluid package
 *
 * COPYRIGHT NOTICE
 * Copyright (c) 2007 Herbert H.H. Chang, Daniel J. Valentino, Gary R. Duckwiler, and Arthur W. Toga
 * Laboratory of Neuro Imaging, Department of Neurology, UCLA.
 */
 
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.GregorianCalendar;
import java.util.Date;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Polygon;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JButton;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;
import javax.swing.JCheckBox;
import javax.swing.SwingUtilities;
import javax.swing.border.TitledBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

import java.awt.image.BufferedImage;
import java.awt.Graphics2D;
import java.awt.AlphaComposite;

import org.medtoolbox.jviewbox.viewport.ViewportGrid;
import org.medtoolbox.jviewbox.viewport.Viewport;
import org.medtoolbox.jviewbox.viewport.annotation.Annotation;
import org.medtoolbox.jviewbox.viewport.annotation.DynamicAnnotationShape;

import tools.Matrix;
import tools.Poisson;


/**
 * The <code>JControlSet</code> class is dedicated to the GUI setting for controlling 
 * parameters used in the evolution process and the charged fluid algorithm.
 */
public class JControlSet extends JPanel
{
	/**
	 * The <code>ChargedFluidComputation</code> class is an inner class for handling  
	 * the charged fluid evolution.
	 */
  	private class ChargedFluidComputation extends Thread
  	{
  		/** 
  		 * The hour of the computation time. 
  		 */
		private int _hour;
		
		/** 
		 * The minute of the computation time. 
		 */
		private int _min;
		
		/** 
		 * The second of the computation time. 
		 */
		private int _sec;
		
		/** 
		 * The mini second of the computation time. 
		 */
		private int _msec;
		
  		
  		/** 
		 * Performs the charged fluid evolution and computation .
		 */
		public void run()
		{
			try {
				// Set the starting time.
				GregorianCalendar gc = new GregorianCalendar();
				Date startDate = gc.getTime();
				long startMS = startDate.getTime();
				
				// Perform the initialize displacement of the Charged Fluid.
				_matrix.computeInitialParameter();
						
				// Do the following numbers of iteration.
				int iter = 0;
				boolean isAllPoissonFrozen = false;
				while (!isAllPoissonFrozen) {
					iter++;
					// Update the Charged Fluid element's positions
					_matrix.updateChargedFluid();
					
					// Use the swing utilities to perform the viewport repaint such that the
					// repaint is executed in the swing to prevent inconsistent thread warnings.					
					Runnable repaintViewport = new Runnable() {
						public void run() {
							// Repaint the view port for display.
							_grid.repaint(_origViewport);
						}
					};				
					SwingUtilities.invokeLater(repaintViewport);
				
					// For pausing the iteration.
					int pauseInt = _matrix.getPause();
					int modInt = (int)(_matrix.getPause() / 2);
					// While it is odd, pause the program.
					while ((pauseInt - modInt * 2) == 1) {
						pauseInt = _matrix.getPause();
						modInt = (int)(_matrix.getPause() / 2);
					}

					if (_matrix.isBreak()) {
						System.out.print("\n*** Terminated by users ***");
						break;
					}
									
					// Check if the stop criteria are satisfied.
					isAllPoissonFrozen = _matrix.isAllPoissonFrozen();
				}
						
				// Trace the charged fluid front representing the contour.
				_matrix.traceChargedFluidFront();
				
				// Set the ending time.
				gc = new GregorianCalendar();
				Date endDate = gc.getTime();
				long endMS = endDate.getTime();
				
				// Add the final CF to the vector. 
				_matrix.addCF2Vector();
				
				System.out.print("\nEnd Iteration.");
				
				// Repaint the viewport.
				_matrix.showCFEvolutionContours();
				_grid.repaint(_origViewport);
				
				// Compute the elapsed time.
				long elapsedMS = endMS - startMS;
				// Compute the corresponding hour, minute, and second.
				_computeHourMinSecond(elapsedMS);
				
				String strg = "Computation time: ";
				if (_hour > 0) {
					Integer hourInt = new Integer(_hour);
					strg += hourInt.toString() + " Hour, ";
				}
				if (_min > 0) {
					Integer minInt = new Integer(_min);
					strg += minInt.toString() + " Min, ";
				}
				if (_sec > 0) {
					Integer secInt = new Integer(_sec);
					strg += secInt.toString() + " Sec, ";
				}
				Integer msecInt = new Integer(_msec);
				strg += msecInt.toString() + " MiniSec";
				System.out.print("\n" + strg);
				// Set the computation time to the Matrix.
				_matrix.timeStrg = strg;
			}
			catch (Exception e){
				System.out.println("Error:  " + e.getMessage());	
			}	
		}
		
		
		/**
		 * Calculates the computation time in terms of hour, min, sec and minisec with the
		 * given time in minisecs.
		 * @param	ms	computation time in minisecond.
		 */
		private void _computeHourMinSecond(long ms)
		{
			int timeInSec = (int)(ms/1000);
			_msec = (int)(ms - timeInSec * 1000);
			int timeInMin = (int)(timeInSec/60);
			_sec = timeInSec - timeInMin * 60;
			_hour = (int)(timeInMin/60);
			_min = timeInMin - _hour * 60;
			
		}
  	}
	
	/**
	 * Arranging the display layout of the input images.
	 */
	private ViewportGrid _grid;
	
	/**
	 * The original view port for the image display.
	 */
	private Viewport _origViewport;
	
	/** 
	 * The original buffered image corresponding to the <tt>_origViewport</tt>. 
	 */
  	private BufferedImage _origImage;
  	
	/** 
     * A special class for the input image computation, display and manipulation. 
     */
    private Matrix _matrix;
    
	/** 
  	 * Weighting factor for the electric potential to balance the image potential, which 
  	 * is the modulus of the image gradient. 
  	 */
	private double _beta;
	
	/**
	 * The specified electrostatic equilibrium condition for the system. 
	 */
	private double _gamma;
	
	/** 
	 * A text field Swing component for the value of beta.
	 */
	private JTextField _betaText = new JTextField(3);
	
	/** 
	 * A slider Swing component for the integer value of beta.
	 */
	private JSlider _betaSlider;
	
	/** 
	 * A slider Swing component for the decimal value of beta.
	 */
	private JSlider _betaSlider2;
	
	/** 
	 * A radio button Swing component of using circle-like shapes for the initial contours.
	 */
	private JRadioButton _circleButton;
	
	/** 
	 * A radio button Swing component of using square shapes for the initial contours.
	 */
	private JRadioButton _squareButton;
	
	/** 
	 * A radio button Swing component for using filter #1.
	 */
	private JRadioButton _filter1Button;
	
	/** 
	 * A radio button Swing component for using filter #2.
	 */
	private JRadioButton _filter2Button;
	
	/** 
	 * A button Swing component for submitting the parameter changes.
	 */
	private JButton _actionSubmitButton = new JButton("Submit");
	
	/** 
	 * A button Swing component for pausing the program.
	 */
	private JButton _pauseButton = new JButton("Pause");
	
	/** 
	 * A button Swing component for starting the program.
	 */
	private JButton _startButton = new JButton("Start");
	
	/** 
	 * A button Swing component for terminating the program.
	 */
	private JButton _terminateButton = new JButton("Terminate");
	
	
	/**
	 * Constructs a new control set Swing component with the specified <code>Matrix</code>,
	 * <code>ViewportGrid</code>, and <code>JFilterSet</code>.
	 * @param	matrix				a special class for handling the input image, display,
	 *								and manipulation.
	 * @param	grid				a view port grid containing the input images.
	 */
	public JControlSet(Matrix matrix, ViewportGrid grid)
	{
		_matrix = matrix;
		_grid = grid;
		_beta = _matrix.getBeta();
		_gamma = _matrix.getGamma();
		Iterator it = _grid.getViewports().iterator();
		_origViewport = (Viewport)it.next();
		_origViewport = (Viewport)it.next();
		_origImage = (BufferedImage)_origViewport.getImage();
		
		_circleButton = new JRadioButton("Circle", false);
		_squareButton = new JRadioButton("Square", true);
		_filter1Button = new JRadioButton("#1: 3x3", false);
		_filter2Button = new JRadioButton("#2: 5x5", false);
		int indexOfFilter = _matrix.getIndexOfFilter();
		if (indexOfFilter == 1) {
			_filter1Button.setSelected(true);
		}
		else if (indexOfFilter == 2) {
			_filter2Button.setSelected(true);
		}
		
		// Parameter initialization.
		this.initialParameter();
		// Setup of GUI.
		this.setupOfGUI();
		// Add the listeners to the control elements.
		this.addListener();
	}
	
	
	/**
	 * Initializes and allocates the parameters for display.
	 */
	public void initialParameter()
	{
		_betaSlider = new JSlider(0, 30, (int)_beta);
		_betaSlider2 = new JSlider(0, 9, (int)((_beta - (int)_beta + 0.05)*10));
		
		// JTextField initialization.
		Integer value = new Integer(_betaSlider.getValue()*10 + _betaSlider2.getValue());
		Double valued = new Double(value.doubleValue()/10.0);
		_betaText.setText(valued.toString());
	}
	
	
	/**
	 * Adds listeners to the Swing components subject to the changes of status.
	 */
	public void addListener()
	{
		/* Beta */
		// Add ChangeListener to the JSlider subjected to stateChanged.
		_betaSlider.addChangeListener(
			new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					Integer beta = new Integer(_betaSlider.getValue()*10 + _betaSlider2.getValue());
					Double valued = new Double(beta.doubleValue()/10.0);
					_betaText.setText(valued.toString());	
					_betaText.setForeground(Color.red);
				}
			}
		);
		// Add ChangeListener to the JSlider subjected to stateChanged.
		_betaSlider2.addChangeListener(
			new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					Integer beta = new Integer(_betaSlider.getValue()*10 + _betaSlider2.getValue());
					Double valued = new Double(beta.doubleValue()/10.0);
					_betaText.setText(valued.toString());	
					_betaText.setForeground(Color.red);
				}
			}
		);
		// Add ActionListener to the JTextField responding to the JSlider
		_betaText.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					_betaText.setForeground(Color.red);
					Double betad = new Double(_betaText.getText());
					int betai = (int)betad.doubleValue();
					_betaSlider.setValue(betai);
					_betaSlider2.setValue((int)((betad.doubleValue() - (double)betai + 0.05)*10));
				}
			}
		);
		// Add ChangeListener to the JRadioButtons subjected to stateChanged.
		_circleButton.addChangeListener(
			new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					if (_circleButton.isSelected()) {
						_matrix.setSquare(false);
					}
					else {
						_matrix.setSquare(true);
					}
				}
			}
		);
		_filter1Button.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					_matrix.setIndexOfFilter(1);
				}
			}
		);
		_filter2Button.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					_matrix.setIndexOfFilter(2);
				}
			}
		);
		
		// Add ActionListener to the action JButtons.
		_actionSubmitButton.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					Double betaDouble = new Double(_betaText.getText());
					_matrix.setBeta(betaDouble.doubleValue());
					// Set the TextField color blue.
					_betaText.setForeground(Color.blue);
				}
			}
		);
		_pauseButton.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					_matrix.setPause();
				}
			}
		);
		_startButton.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					_matrix.setBreak(false);
					_startChargedFluid();
				}
			}
		);
		_terminateButton.addActionListener(
			new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					_matrix.setBreak(true);
				}
			}
		);
	}
	
	
	/**
	 * Creates a new GUI for showing the Swing components.
	 */
	public void setupOfGUI()
	{
		// Setup of JSlider parameters.
		_betaSlider.setPaintTicks(true);
		_betaSlider.setMajorTickSpacing(10);
		_betaSlider.setMinorTickSpacing(2);
		_betaSlider.setPaintLabels(true);
		_betaSlider2.setPaintTicks(true);
		_betaSlider2.setMajorTickSpacing(1);
		_betaSlider2.setPaintLabels(true);
    	
    	// Create logical relationship between radiobuttons
		ButtonGroup radioFilterGroup = new ButtonGroup();
    	radioFilterGroup.add(_filter1Button);
    	radioFilterGroup.add(_filter2Button);
    	ButtonGroup shapeGroup = new ButtonGroup();
    	shapeGroup.add(_circleButton);
    	shapeGroup.add(_squareButton);
    	
    	// Set TextField color blue.
    	_betaText.setForeground(Color.blue);
    	
		// Setup of Panel arrangement
		JPanel betaSliderPanel = new JPanel(new BorderLayout());
		betaSliderPanel.add(_betaSlider, BorderLayout.NORTH);
		betaSliderPanel.add(_betaSlider2, BorderLayout.SOUTH);
		
		JPanel betaPanel = new JPanel(new BorderLayout());
		JLabel betaLabel = new JLabel("Beta: ");
		betaPanel.add(betaLabel, BorderLayout.WEST);
		betaPanel.add(_betaText, BorderLayout.EAST);
		betaPanel.add(betaSliderPanel, BorderLayout.SOUTH);
		
		JPanel shapePanel = new JPanel(new BorderLayout());
		JLabel shapeLabel = new JLabel("Contour shape: ");
		shapePanel.add(shapeLabel, BorderLayout.NORTH);
		shapePanel.add(_circleButton, BorderLayout.WEST);
		shapePanel.add(_squareButton, BorderLayout.CENTER);
		
		JPanel filterPanel = new JPanel(new BorderLayout());
		JLabel filterLabel = new JLabel("Filter kernel: ");
		filterPanel.add(filterLabel, BorderLayout.NORTH);
		filterPanel.add(_filter1Button, BorderLayout.WEST);
		filterPanel.add(_filter2Button, BorderLayout.CENTER);
		
		JPanel actionButtonPanel = new JPanel(new GridLayout(2, 2, 5, 5));	
		actionButtonPanel.add(_actionSubmitButton);
		actionButtonPanel.add(_pauseButton);
		actionButtonPanel.add(_startButton);
		actionButtonPanel.add(_terminateButton);
		
		// Main panels : _paraPanel, _displayPanel, and _actionPanel
		JPanel paraPanel = new JPanel(new BorderLayout(0, 5));
		paraPanel.add(shapePanel, BorderLayout.NORTH);
		paraPanel.add(betaPanel, BorderLayout.CENTER); 
		paraPanel.add(filterPanel, BorderLayout.SOUTH);
		paraPanel.setBorder(new TitledBorder(new EtchedBorder(), "PARAMETER"));

		JPanel actionPanel = new JPanel(new BorderLayout(0, 5));
		actionPanel.add(actionButtonPanel, BorderLayout.CENTER);
		actionPanel.setBorder(new TitledBorder(new EtchedBorder(), "ACTION"));
		
		// Add main panels to be shown.
		int width = 100;
		this.setSize(width, 2*width);
		this.setLayout(new BorderLayout(0,10));
		this.add(paraPanel, BorderLayout.NORTH);
		this.add(actionPanel, BorderLayout.SOUTH);	
	}
	
	
	/**
	 * Starts the charged fluid algorithm by threading the inner class, 
	 * <code>ChargedFluidComputation</code>.
	 */
	private void _startChargedFluid()
	{
		int width = _origImage.getWidth();
      	int height = _origImage.getHeight();
      	/* Create a buffered image for displaying the charge density. */
    	BufferedImage startImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
    	/* Set the RGB buffered image. */
    	int showRGB = 255 << 8;		//green color
    	for (int y = 0; y < height; y++)
    	{
      	  for (int x = 0; x < width; x++)
          {
            startImage.setRGB(x, y, showRGB);
          }
        }
        Graphics2D displayGraphics = _origImage.createGraphics();
    	displayGraphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.3f));
    	displayGraphics.drawImage(startImage, null, 0, 0);
    	// Repaint the Viewports
        _grid.repaint(_origViewport);
        // Start the algorithm    
    	ChargedFluidComputation cfc = new ChargedFluidComputation();
    	cfc.start();	
	}
}