/*
 * @(#)Viewer.java		1.10 07/01/20
 *
 * ChargedFluid package
 *
 * COPYRIGHT NOTICE
 * Copyright (c) 2007 Herbert H.H. Chang, Daniel J. Valentino, Gary R. Duckwiler, and Arthur W. Toga
 * Laboratory of Neuro Imaging, Department of Neurology, UCLA.
 */
 
import java.awt.BorderLayout;
import java.awt.Point;
import java.awt.Font;
import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import javax.swing.JFileChooser;
import javax.swing.JColorChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.stream.FileImageInputStream;

import org.medtoolbox.jviewbox.imagesource.ImageReaderAdapter;
import org.medtoolbox.jviewbox.viewport.ViewportGrid;
import org.medtoolbox.jviewbox.viewport.ViewportToolBar;
import org.medtoolbox.jviewbox.viewport.Viewport;
import org.medtoolbox.jviewbox.viewport.annotation.StaticAnnotationString;

import tools.Matrix;
import tools.PanTool;
import tools.WindowLevelTool;
import tools.ZoomTool2;
import tools.SelectTool;
import tools.ChargedFluidTool;


/**
 * The <code>Viewer</code> class is dedicated to display a set of medical images. The viewer
 * reads images from the file names provided on the command line as arguments
 * and displays them.
 */
public class Viewer extends JFrame
{
	/**
	 * Storing the input images from the original data.
	 */
	private Vector _allImages = new Vector(2);
	
	/**
	 * Arranging the display layout of the input images.
	 */
	private ViewportGrid _grid;
	
	/**
	 * A typical <code>JToolBar</code> for managing the function tools to the images.
	 */
	private ViewportToolBar _toolBar;
	
	/** 
	 * A typical <code>JPanel</code> for controlling the parameters during the evolution
	 * process and managing the image and contour display. 
	 */
	private JControlSet _jControlSet;
	
	/** 
	 * The default color for the contour. 
	 */
	private Color _contourColor = Color.yellow;
	
	/** 
	 * Input image width. 
	 */
	private int _width;
	
	/** 
	 * Input image height. 
	 */
    private int _height;
    
    /** 
     * Integer parameters for the evolution process: index of power and index of filter.
     */
    private int[] _intParameter = {1, 1};
    
    /** 
     * Double parameters for the evolution process: beta. 
     */
    private double[] _parameter = {1.2};
    
    /**
	 * Boolean for indicating if the result is superimposed by the input image for display,
	 * if displaying the intensity image, if the gamma value is normal or not, 
	 * and if the image gradient is used as the effective field.
	 */
    private boolean[] _boolean = {true, true, true, false};
    
    /** 
     * A special class for the input image computation, display and manipulation. 
     */
  	private Matrix _matrix = new Matrix(_intParameter, _parameter, _boolean);
    

    /**
     * Constructs a viewer to display the input images with the specified file.
     * @param	file		a file representing the input image.
     */
    public Viewer(File file)
    {
		super("Charged Fluid Model for brain image segmentation");
		
		// Create the GUI showing the images and toolbars.
		_setGUI(file);	
	 }


    /**
     * Application main method to start the whole process.
     * @param	args	command-line arguments as a image file name to view.
     */
    public static void main(String[] args) 
	{
		File file;
		// If there is no command-line arguments, pop up a JFileChooser and
		// let the user choose image file to load
		if (args.length == 0) {
	    	JFileChooser chooser = new JFileChooser("./");
	    	chooser.setFileHidingEnabled(true);
	    	chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
	    	if (chooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
				file = chooser.getSelectedFile();
	    	}
	    	else {
	    		file = null;
	 		}
		}
		// Use command-line arguments
		else {
			file = new File(args[0]);
		}
		
		// Create a Viewer and make it visible
		Viewer viewer = new Viewer(file);
		viewer.setDefaultCloseOperation(EXIT_ON_CLOSE);
		viewer.setSize(800, 600);
		viewer.setVisible(true);
    }


	/**
	 * Reads the image file using the image I/O plugins.
	 * @param	file	an image file that are selected.
	 */
	public void _imageIO(File file)
	{
		try {
			// 1st try: use FileImageInputStream as input
			Object input = new FileImageInputStream(file);
			Iterator it = ImageIO.getImageReaders(input);
			// Found no reader at the 1st try?
			if (!it.hasNext()) {
			    // 2nd try: use the file itself as input
			    input = file;
			    it = ImageIO.getImageReaders(file);
			    // Found no reader at the 2nd try?
			    if (!it.hasNext()) {
					// 3rd try: get a reader from the file name suffix
					String fileName = file.getName();
					int lastDot = fileName.lastIndexOf('.');
					if (lastDot >= 0) {
				    	String suffix = fileName.substring(lastDot + 1);
				    	it = ImageIO.getImageReadersBySuffix(suffix);
					}
				}
			}
			// Found any reader at all?
			if (it.hasNext()) {
			    ImageReader reader = (ImageReader)it.next();
			    reader.setInput(input, false);
				// Get the width and height of the first image.
				_width = reader.getWidth(0);
				_height = reader.getHeight(0);
			    // Wrap each image in the reader in an adapter for jViewBox
			    int numImages = reader.getNumImages(true);
			    for (int j = 0; j < numImages; j++) {
					_allImages.add(new ImageReaderAdapter(reader, j));
			    }
			}
			// No reader available for the file
			else {
			    System.out.println("Unable to find an image decoder for " + file);
			    System.out.println("Skipping the file");
			}
		}
		catch (IOException e) {
			System.out.println("Unable to read " + file + " due to " + e);
			System.out.println("Skipping the file");
		}
		// Check for at least one image to show
		if (_allImages.isEmpty()) {
		    System.out.println("No image to show");
		    System.exit(1);
		} 	
	}
  
  
	/**
	 * Adds tools to the toolbar.
	 * @param	file	the input file to be referred for some tools.
	 */
	private void _addToolBar(File file)
	{
		_toolBar = new ViewportToolBar();
		// Add tools.
		_toolBar.addTool(new PanTool(), "Pan");
		_toolBar.addTool(new ZoomTool2(), "Zoom");
		_toolBar.addTool(new WindowLevelTool(), "Window/Level");
		_toolBar.addTool(new SelectTool(_matrix), "Select");
		_toolBar.addTool(new ChargedFluidTool(_matrix), "CFM");
	}


	/**
	 * Creates a blank image for the duplicate image in RGB mode and generates a GUI
	 * for the manipulation of the program.
	 * @param	file	the input file to be referred.
	 */
	private void _setGUI(File file)
	{
		if (file == null) {
			return;
		}
		// Read image via function _imageIO
		_imageIO(file);	
		
		// Set the image width and height.
		_matrix.setImageWidth(_width);
		_matrix.setImageHeight(_height);
		
		// Create a ToolBar and add Tools to it
		_addToolBar(file);
		
		// Generate a image in ARGB model
		BufferedImage blankImage = new BufferedImage(_width, _height, BufferedImage.TYPE_INT_ARGB);	
		_allImages.add(blankImage);
		
		// Create a ViewportGrid to display the images
		_grid = new ViewportGrid(_allImages);
		// Add annotation string to the original viewport
		Iterator it = _grid.getViewports().iterator();
		Viewport origViewport = (Viewport)it.next();
		StaticAnnotationString saString = new StaticAnnotationString("ORIGIN",
						 new Point(10,10), new Font(null, Font.ITALIC, 38));
		saString.setForegroundColor(Color.red);
		saString.setBackgroundColor(Color.blue);
		origViewport.addAnnotation(saString);
		
		// Link the ToolBar to the ViewportGrid
		_grid.registerViewportToolBar(_toolBar);
		// Create a JControlSet for adjusting controlled parameters.
		_jControlSet = new JControlSet(_matrix, _grid);
		// Remove all container elements.
		getContentPane().removeAll();
		// Arrange the ToolBar and ViewportGrid in this frame
		getContentPane().add(_toolBar, BorderLayout.NORTH);
		getContentPane().add(_grid, BorderLayout.CENTER);
		getContentPane().add(_jControlSet, BorderLayout.WEST);
		getContentPane().add(new JLabel(file.toString()), BorderLayout.SOUTH);	
	}
}
