/*
 * @(#)Poisson.java		1.10 07/01/20
 *
 * ChargedFluid package
 *
 * COPYRIGHT NOTICE
 * Copyright (c) 2007 Herbert H.H. Chang, Daniel J. Valentino, Gary R. Duckwiler, and Arthur W. Toga
 * Laboratory of Neuro Imaging, Department of Neurology, UCLA.
 */
 
package tools;

import java.awt.Point;
import java.util.Vector;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * The <code>Poisson</code> class is dedicated to the electric potential computation
 * using Poisson's equation for the charged fluid algorithm. There is only one free
 * <code>ChargedFluid</code> during the evolution process, which is stored in a linkedlist.
 * After the evolution is over, a contour tracing technique is used to create the final
 * charged fluids, whose number is at least one. If there are several disconnected charged
 * fluids that are not traced in this Poisson domain, the system will trace all of them.
 * The dimension of the computation area is automatically adjusted to fit the need of the
 * FFT calculation based upon the sizes of the free charged fluid.
 */
public class Poisson
{
	/** 
	 * Storing the <code>ChargedFluid</code> class. 
	 */
	public LinkedList linkedList;
	
  	/** 
  	 * Weighting factor for the electric potential to balance the image potential, which 
  	 * is the modulus of the image gradient. 
  	 */
	private double _beta;
  	
	/** 
	 * The minimum position of the charged fluid in x-axis. 
	 */
	public int minXAxis;
	
	/** 
	 * The maximum position of the charged fluid in x-axis. 
	 */
	public int maxXAxis;
	
	/** 
	 * The minimum position of the charged fluid in y-axis. 
	 */
	public int minYAxis;
	
	/** 
	 * The maximum position of the charged fluid in y-axis. 
	 */
	public int maxYAxis;
	
	/** 
	 * The width of the electric field and potential arrays for the computation (x-axis). 
	 */
	private int _arrayWidth;
	
	/** 
	 * The height of the electric field and potential arrays for the computation (y-axis). 
	 */
	private int _arrayHeight;
	
	/** 
	 * The origin coordinate in x-axis of the Poisson w.r.t the Matrix domain. 
	 */
	private int _xOrigin;
	
	/** 
	 * The origin coordinate in y-axis of the Poisson w.r.t the Matrix domain. 
	 */
	private int _yOrigin;
	
	/** 
	 * The starting coordinate in x-axis w.r.t the Poisson domain for the array computation. 
	 */
	private int _xStart;
	
	/** 
	 * The starting coordinate in y-axis w.r.t the Poisson domain for the array computation. 
	 */
	private int _yStart;
	
	/** 
	 * The ending coordinate in x-axis w.r.t the Poisson domain for the array computation. 
	 */
	private int _xEnd;
	
	/** 
	 * The ending coordinate in y-axis w.r.t the Poisson domain for the array computation. 
	 */
	private int _yEnd;
	
	/** 
	 * The start position for the contour tracing. 
	 */
  	private Point _startPosition;
	
	/** 
	 * Temporarily storing the charges after the contour tracing procedure.
	 */
  	private LinkedList _newChargedFluidLink;
	
	/** 
	 * Image intensity array from the Matrix domain. 
	 */
	private double[][] _imageIntensityArray;
	
	/** 
	 * Image gradient array from the Matrix domain. 
	 */
	private double[][] _imageGradientArray;
	
	/**
	 * Image gradient array in x-axis. 
	 */
	private double[][] _gradientArrayInX;
	
	/**
	 * Image gradient array in y-axis. 
	 */
	private double[][] _gradientArrayInY;
	
	/** 
	 * Indicating the index of the ChargedFluid for the 2-pixel-wide interface. 
	 */
  	private int[][] _indexArray;
	
	/** 
	 * Electric potential array, also the real part for the FFT computation.
	 */
  	private double[][] _electricPotentialArray;
  	
  	/** 
  	 * Charge array #1 for storing the charges to alternate the charges
  	 * in the charge distribution procedure. 
  	 */
  	private double[][] _chargeArrayOne;

  	/** 
  	 * Charge array #2 for storing the charges to alternate the charges
  	 * in the charge distribution procedure. 
  	 */
  	private double[][] _chargeArrayTwo;
  	
  	/** 
  	 * Referring to charge array #1 or #2 in the front and contour computation procedure. 
  	 */
  	private double[][] _chargeArray;
  	
  	/** 
  	 * Electric field array in x-axis. 
  	 */
	private double[][] _electricFieldArrayInX;
	
	/** 
	 * Electric field array in y-axis, also the imaginary part for the FFT computation. 
	 */
	private double[][] _electricFieldArrayInY;
  	
  	/** 
  	 * True, if the position is occupies by a free charge.
  	 */
  	private boolean[][] _freeChargeArray;
  	
  	/** 
  	 * Boolean array for indicating the free charged fluid trace. True, if the position is 
  	 * passed by the free charged fluid. T:(+) passed and F:(-) empty. The update of this
  	 * array is performed for each individual fluid element right after it deforms the shape.
  	 */
  	private boolean[][] _traceBooleanArray;
  	
  	/** 
  	 * Boolean array for indicating the free charged fluid trace. True, if the position is 
  	 * passed by the free charged fluid. T:(+) passed and F:(-) empty. The update of this
  	 * array is performed after all charged fluid elements have deformed their shape.
  	 */
  	private boolean[][] _labelBooleanArray;
  	
  	/** 
  	 * A constant to normalize the electric potential of the charge for multiple 
  	 * charged fluids. 
  	 */
  	private double _normalizedPotential;
  	
	/** 
	 * The last maximum charge used to limit the next largest charge point.
	 */
	private double _lastMaxCharge;
	
  	/**
	 * True, if the image gradient is directly used as the effective field.
	 */
	private boolean _isGradientField;
	
  	
	/** 
	 * Constructs a poisson with the specified charged fluid list, the trace boolean array,
	 * the image intensity, the image gradient array, the potential weight (beta),
	 * the normalized potential constant, the range allowed for the mean intensity,
	 * and the boolean of using the extra charge arrangement or not. 
	 *
	 * @param	linkedList				a list contains the charged fluid class.
	 * @param	traceBooleanArray		the boolean array from the Matrix domain. True, if
	 *									the position is passed by the free charged fluid.
	 * @param 	labelBooleanArray		Similar to the <tt>traceBooleanArray</tt> except the
	 *									update is performed after all charged fluids are deformed.
	 * @param	imageIntensityArray		the image intensity array from the Matrix domain.
	 * @param	imageGradientArray		the image gradient magnitude array from the the Matrix domain.
	 * @param	gradientArrayInX		the image gradient in x-axis from the Matrix domain.
	 * @param	gradientArrayInY		the image gradient in y-axis from the Matrix domain.
	 * @param	beta					the weight for balancing the electric potential and
	 *									the image potential.
	 * @param	normalizedPotential		a constant to normalize the electric potential for 
	 *									multiple charged fluids.
	 * @param	isGradientField			a boolean that determines whether the image gradient is
	 *									directly used as the effective field or not.
	 */
	public Poisson(LinkedList linkedList, boolean[][] traceBooleanArray, boolean[][] labelBooleanArray,
		double[][] imageIntensityArray, double[][] imageGradientArray, double[][] gradientArrayInX, double[][] gradientArrayInY, 
		double beta, double normalizedPotential, boolean isGradientField)
	{
		this.linkedList = linkedList;
		_traceBooleanArray = traceBooleanArray;
		_labelBooleanArray = labelBooleanArray;
		_imageIntensityArray = imageIntensityArray;
		_imageGradientArray = imageGradientArray;
		_gradientArrayInX = gradientArrayInX;
		_gradientArrayInY = gradientArrayInY;
		_beta = beta;
		_normalizedPotential = normalizedPotential;
		_isGradientField = isGradientField;
	}
	

	/** 
	 * Calculates the origin and size of the computation arrays. 
	 */
	public void computePotentialArrayDimension()
	{
		int xComputeRange = maxXAxis - minXAxis + 1;
		int yComputeRange = maxYAxis - minYAxis + 1;
		if ((xComputeRange > 0) && (yComputeRange > 0)){
			_arrayWidth = _getOptimalFFTLength(xComputeRange);
			_arrayHeight = _getOptimalFFTLength(yComputeRange);
			_xOrigin = (int)(maxXAxis + minXAxis - _arrayWidth) / 2 + 1;
			_yOrigin = (int)(maxYAxis + minYAxis - _arrayHeight) / 2 + 1;
			_xStart = minXAxis - 2 - _xOrigin;
			_yStart = minYAxis - 2 - _yOrigin;
			_xEnd = maxXAxis + 3 - _xOrigin;
			_yEnd = maxYAxis + 3 - _yOrigin;
		}
		else {
			System.out.println(" Potential Array Dimension Computation Error: incorrect size");	
		}
	}
	
	
	/** 
	 * Returns the smallest integer power of 2 that is larger than the given length.
	 * The computation is based upon the safe distance of using
	 * the periodic boundary to solve Poisson's equation.
	 * @param	length	the length to be used for the computation.
	 * @return	the optimal length for the FFT computation.
	 */
	private int _getOptimalFFTLength(int length)
	{
		if (length < 12) {
			return 16;
		}
		else if (length < 24) {
			return 32;
		}
		else if (length < 48) {
			return 64;
		}
		else if (length < 98) {
			return 128;
		}
		else if (length < 194) {
			return 256;
		}		
		else if (length < 386) {
			return 512;
		}
		else {
			length *= 1.35;
			double powerOfLength = Math.ceil(Math.log(length) / Math.log(2.0));
			return (int)Math.pow(2.0, powerOfLength);
		}				
	}
	
	
	/** 
	 * Calculates the computation array size and allocates the arrays.
	 */
	public void initialization()
	{
		computePotentialArrayDimension();
		_electricPotentialArray = new double[_arrayHeight][_arrayWidth];
		_chargeArrayOne = new double[_arrayHeight][_arrayWidth];
    	_chargeArrayTwo = new double[_arrayHeight][_arrayWidth];
    	_electricFieldArrayInX = new double[_arrayHeight][_arrayWidth];
    	_electricFieldArrayInY = new double[_arrayHeight][_arrayWidth];
    	_indexArray = new int[_arrayHeight][_arrayWidth];
    	_freeChargeArray = new boolean[_arrayHeight][_arrayWidth];
	}
	
	
	/** 
	 * Computes the initial electric fields of the charges for the very beginning front
	 * deformation procedure.
	 */
	public void computeInitialElectricField()
	{
		for (int y = 0; y < _arrayHeight; y++) {
			for (int x = 0; x < _arrayWidth; x++) {
				_indexArray[y][x] = -1;
			}
		}
		Iterator itf = linkedList.iterator();
		while (itf.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)itf.next();
		  	int indexOfCF = linkedList.indexOf(currentCF);
		  	Iterator itcf = currentCF.linkedList.iterator();
		  	while (itcf.hasNext()) {
		    	Charge currentCharge = (Charge)itcf.next();
		    	Point chargePt = currentCharge.position;
		    	int xPoisson = chargePt.x - _xOrigin;
		    	int yPoisson = chargePt.y - _yOrigin;
		    	_indexArray[yPoisson][xPoisson] = indexOfCF;
		    }
		}
		_setCharge2ArrayOne();
		_setElectricPotentialArray(_chargeArrayOne);
		_computeElectricPotentialArray();
		double maxField = _getMaxElectricField();
		_chargeInterpolation(_chargeArrayOne, _chargeArrayTwo, maxField);
		_chargeInterpolation(_chargeArrayTwo, _chargeArrayOne, maxField);
		_setElectricPotentialArray(_chargeArrayOne);
		_computeElectricPotentialArray();
		
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)it.next();
		  	Iterator itc = currentCF.linkedList.iterator();
		  	while (itc.hasNext()) {
		    	Charge currentCharge = (Charge)itc.next();
		    	Point chargePt = currentCharge.position;
		    	int x = chargePt.x - _xOrigin;
		    	int y = chargePt.y - _yOrigin;
		        int xm1 = x - 1;
		        int ym1 = y - 1;
		        int xp1 = x + 1;
		        int yp1 = y + 1;
		        double electricPhiXM = _electricPotentialArray[y][xm1];
		        double electricPhiXP = _electricPotentialArray[y][xp1];
		        double electricPhiYM = _electricPotentialArray[ym1][x];
		        double electricPhiYP = _electricPotentialArray[yp1][x];
		        currentCharge.effectiveFieldInX = (electricPhiXM - electricPhiXP);
		        currentCharge.effectiveFieldInY = (electricPhiYM - electricPhiYP);
		    }
		}
	}
	
	
	/**
	 * Returns the origin coordinate in x-axis w.r.t. the Matrix domain.
	 * @return	the origin coordinate in x-axis.
	 */
	public int getXOrigin()
	{
		return _xOrigin;	
	}
	
	
	/**
	 * Returns the origin coordinate in y-axis w.r.t. the Matrix domain.
	 * @return	the origin coordinate in y-axis.
	 */
	public int getYOrigin()
	{
		return _yOrigin;	
	}
	
	
	/**
	 * Returns the width of the computation array.
	 * @return	the value of the array width.
	 */
	public int getArrayWidth()
	{
		return _arrayWidth;	
	}
	
	
	/**
	 * Returns the height of the computation array.
	 * @return	the value of the array height.
	 */
	public int getArrayHeight()
	{
		return _arrayHeight;	
	}
	
	
	/** 
	 * Returns the width (x-axis) of the poisson.
	 * @return	the value of the poisson width.
	 */
	public int getWidth()
	{
		return (maxXAxis - minXAxis + 1);
	}
	
	
	/** 
	 * Returns the height (y-axis) of the poisson.
	 * @return	the value of the poisson height.
	 */
	public int getHeight()
	{
		return (maxYAxis - minYAxis + 1);
	}
	
	
	/** 
	 * Deforms the 1-pixel-wide contour to the 2-pixel-wide interface based on the effective
	 * field directions of charges in the free charged fluid. The nearest four neighbors of
	 * the charges are marked using an index array and a trace boolean array.
	 */ 
	public void frontDeformation()
	{
		for (int y = 0; y < _arrayHeight; y++) {
			for (int x = 0; x < _arrayWidth; x++) {
				_indexArray[y][x] = -1;
			}
		}
		Iterator itv = linkedList.iterator();
		while (itv.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)itv.next();
	      	int indexOfCF = linkedList.indexOf(currentCF);
	      	Iterator it = currentCF.linkedList.iterator();
	      	while (it.hasNext()) {
	        	Charge currentCharge = (Charge)it.next();
	        	int x = currentCharge.position.x;
	        	int y = currentCharge.position.y;
	          	double effectiveFieldInX = currentCharge.effectiveFieldInX;
	          	double effectiveFieldInY = currentCharge.effectiveFieldInY;
	          	int xPoisson = x - _xOrigin;
	          	int yPoisson = y - _yOrigin;
	          	int xn, yn;
	          	if (currentCharge.isEffectiveDirectionInverse) {
		          	if (effectiveFieldInX < 0.0) {
		            	xn = xPoisson + 1;
		          	}
		          	else {
		            	xn = xPoisson - 1;
		          	}
		          	if (effectiveFieldInY < 0.0) {
		            	yn = yPoisson + 1;
		          	}
		          	else {
		            	yn = yPoisson - 1;
		          	}
		        }
		        else {
		          	if (effectiveFieldInX >= 0.0) {
		            	xn = xPoisson + 1;
		          	}
		          	else {
		            	xn = xPoisson - 1;
		          	}
		          	if (effectiveFieldInY >= 0.0) {
		            	yn = yPoisson + 1;
		          	}
		          	else {
		            	yn = yPoisson - 1;
		          	}
		        }
	          	if (xn == -1) {
	            	xn = _arrayWidth - 1;
	          	}
	          	if (xn == _arrayWidth) {
	            	xn = 0;
	          	}
	          	if (yn == -1) {
	            	yn = _arrayHeight - 1;
	          	}
	          	if (yn == _arrayHeight) {
	            	yn = 0;
	          	}
	          	_indexArray[yPoisson][xPoisson] = indexOfCF;
	          	_indexArray[yn][xn] = indexOfCF;
	          	if (_labelBooleanArray[yn+_yOrigin][xn+_xOrigin]) {
	          		if (_labelBooleanArray[yPoisson+_yOrigin][xn+_xOrigin]) {
	          			_indexArray[yPoisson][xn] = indexOfCF;
	          		}
	          		if (_labelBooleanArray[yn+_yOrigin][xPoisson+_xOrigin]) {
	          			_indexArray[yn][xPoisson] = indexOfCF;
	          		}
	          	}
	          	else {
		          	_indexArray[yPoisson][xn] = indexOfCF;
		          	_indexArray[yn][xPoisson] = indexOfCF;
		          	_traceBooleanArray[yPoisson+_yOrigin][xn+_xOrigin] = true;
		          	_traceBooleanArray[yn+_yOrigin][xPoisson+_xOrigin] = true;
		          	_traceBooleanArray[yn+_yOrigin][xn+_xOrigin] = true;
	          	}		          	
	      	}
		}
		for (int y = _yStart; y < _yEnd; y++) {
			for (int x = _xStart; x < _xEnd; x++) {
				if (_traceBooleanArray[y+_yOrigin][x+_xOrigin]) {
					_labelBooleanArray[y+_yOrigin][x+_xOrigin] = true;
				}
			}
		}
	}
	
	
	/**
	 * Determines whether the electric direction of the specified charge at the given point
	 * is inward or not based upon the electric fields.
	 * @param	electricFieldInX	the electric field in x-axis.
	 * @param	electricFieldInY	the electric field in y-axis.
	 * @param	pt					the specified <code>Point</code> to be tested.
	 * @return	<code>true</code>, if the the electric field is inward.
	 */
	private boolean _isElectricDirectionInward(double electricFieldInX, double electricFieldInY, Point pt)
	{
		int xMatrix = pt.x;
		int yMatrix = pt.y;
      	int xn;
      	if (electricFieldInX >= 0.0) {
        	xn = xMatrix + 1;
      	}
      	else {
        	xn = xMatrix - 1;
      	}
      	int yn;
      	if (electricFieldInY >= 0.0) {
        	yn = yMatrix + 1;
      	}
      	else {
        	yn = yMatrix - 1;
      	}
      	if (_traceBooleanArray[yMatrix][xn] &&
      		_traceBooleanArray[yn][xMatrix] &&
      		_traceBooleanArray[yn][xn]) {
      		return true;	
      	}
      	else {
      		return false;
      	}
    }
	
	
	/** 
	 * Calculates the overall electric potential from the charge distribution array using
	 * Poisson's equation. The potential is stored in <tt>_electricPotentialArray</tt>.
	 */
	private void _computeElectricPotentialArray()
	{
		for (int y = 0; y < _arrayHeight; y++) {
			for (int x = 0; x < _arrayWidth; x++) {
				_electricFieldArrayInY[y][x] = 0.0;
			}
		}
		FastFourierTransform.fastFT2D(_electricPotentialArray, _electricFieldArrayInY, true);
		for (int v = 0; v < _arrayHeight; v++) {
			for (int u = 0; u < _arrayWidth; u++) {
				if ((v==0) && (u==0)) {
					_electricPotentialArray[v][u] = 0.0;
					_electricFieldArrayInY[v][u] = 0.0;
				}
				else if ((v==_arrayHeight/2) || (u==_arrayWidth/2)) {
					_electricPotentialArray[v][u] = 0.0;
					_electricFieldArrayInY[v][u] = 0.0;
				}
				else {
					double vc, uc;
					if ( v > _arrayHeight/2 ) {
						vc = v - _arrayHeight;
					}
					else {
						vc = v;
					}
					if ( u > _arrayWidth/2 ) {
						uc = u - _arrayWidth;
					}
					else {
						uc = u;
					}
					double coef2 = Math.PI*((vc*vc)/(double)(_arrayHeight*_arrayHeight) + (uc*uc)/(double)(_arrayWidth*_arrayWidth));
					double coef1 = Math.exp(-2.0*Math.PI*coef2);
					_electricPotentialArray[v][u] *= (coef1 / coef2);
					_electricFieldArrayInY[v][u] *= (coef1 / coef2);
				}
			}
		}
		FastFourierTransform.fastFT2D(_electricPotentialArray, _electricFieldArrayInY, false);
	}
	
	
	/** 
	 * Advances the charges in the free charged fluid at the 2-pixel-wide interface
	 * until the specified electrostatic equilibrium condition is achieved, i.e. the system
	 * reaches an electrostatic charge distribution.
	 * @param	gamma		the specified condition for the electrostatic equilibrium.
	 */
	public void chargeDistribution(double gamma)
	{
		_setUniformCharge2ChargedFluid();	
		_setCharge2ArrayOne();
		double ensembleCharge = _getEnsembleCharge();
		double netChargeDifference = ensembleCharge;
		int number = 0;
		while ((netChargeDifference / ensembleCharge) > gamma) {
			_setElectricPotentialArray(_chargeArrayOne);
		  	_computeElectricPotentialArray();
			double maxField = _getMaxElectricField();
	    	_chargeInterpolation(_chargeArrayOne, _chargeArrayTwo, maxField);
	    	_chargeInterpolation(_chargeArrayTwo, _chargeArrayOne, maxField);
			netChargeDifference = _getChargeDifference();
			number++;
		}
	}
	
	
	/** 
	 * Returns the ensemble charge of the electrostatic system in the Poisson domain.
	 * @return	the ensemble charge.
	 */
	private double _getEnsembleCharge()
	{
		double ensembleCharge = 0.0;
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)it.next();
	      	Iterator itc = currentCF.linkedList.iterator();
	      	while (itc.hasNext()) {
	        	Charge currentCharge = (Charge)itc.next();
	        	ensembleCharge += currentCharge.charge;
	        }
		}
		return ensembleCharge;
	}
	
	
	/** 
	 * Returns the pixel by pixel charge difference between <tt>_chargeArrayOne</tt> and 
	 * <tt>_chargeArrayTwo</tt>.
	 * @return	the charge difference.
	 */
	private double _getChargeDifference()
	{
		double sumOfChargeDifference = 0.0;
		for (int y = _yStart; y < _yEnd; y++) {
			for (int x = _xStart; x < _xEnd; x++) {
				if (_indexArray[y][x] >= 0) {
					sumOfChargeDifference += Math.abs(_chargeArrayOne[y][x] - _chargeArrayTwo[y][x]);	
				}
			}
		}
		return sumOfChargeDifference;
	}
	
	
	/** 
	 * Advances the charges from the given charge array one to charge array two using
	 * the Subtracted Dipole Scheme (SUDS) interpolation method at the five discrete 
	 * neighbors. If any of the four neighbors is not inside the spread region (w.r.t. 
	 * _traceBooleanArray), the charge at the position is interpolated to the other charge
	 * positions based upon the corresponding weights.
	 * @param	chargeArrayOne	the given, charge array one (source array).
	 * @param	chargeArrayTwo	the got, charge array two (target array).
	 * @param	maxField		the maximum electric field in the charged fluid.
	 */
	private void _chargeInterpolation(double[][] chargeArrayOne, double[][] chargeArrayTwo, double maxField)
	{
		for (int y = _yStart; y < _yEnd; y++) {
			for (int x = _xStart; x < _xEnd; x++) {
				chargeArrayTwo[y][x] = 0.0;
			}
		}
		for (int y = _yStart; y < _yEnd; y++) {
			for (int x = _xStart; x < _xEnd; x++) {
				double charge = chargeArrayOne[y][x];
				if ((_traceBooleanArray[y+_yOrigin][x+_xOrigin]) && (charge > 0.0)) {
					int xm1 = x - 1;
				    int ym1 = y - 1;
				    int xp1 = x + 1;
				    int yp1 = y + 1;
				  	double dx = 0.5 * _electricFieldArrayInX[y][x] / maxField;
				  	double dy = 0.5 * _electricFieldArrayInY[y][x] / maxField;
				  	double x1 = 0.86602540378 - dx;
				  	double x2 = 0.25 - dx*dx;
					double x3 = 0.86602540378 + dx;
					double x4 = 1.25 - dx*dx;
					double y1 = 0.86602540378 - dy;
					double y2 = 0.25 - dy*dy;
					double y3 = 0.86602540378 + dy;
					double y4 = 1.25 - dy*dy;		  	
					double coeffXYm1 = x4 * y1 * y1 / 4.0;	
					double coeffXm1Y = x1 * x1 * y4 / 4.0;	
					double coeffXY = x2 * y2;				
					double coeffXp1Y = x3 * x3 * y4 / 4.0;	
					double coeffXYp1 = x4 * y3 * y3 / 4.0;	
					double coeffSum = coeffXY;
					double chargeSum = 0.0;
					
					double chargeXYm1 = charge * coeffXYm1;
				  	if (_traceBooleanArray[ym1+_yOrigin][x+_xOrigin]) {
				  		coeffSum += coeffXYm1;
				  	}
				  	else {
				  		chargeSum += chargeXYm1;
				  	}
				  	double chargeXm1Y = charge * coeffXm1Y;
				  	if (_traceBooleanArray[y+_yOrigin][xm1+_xOrigin]) {
				  		coeffSum += coeffXm1Y;
				  	}
				  	else {
				  		chargeSum += chargeXm1Y;
				  	}
				  	double chargeXp1Y = charge * coeffXp1Y;
				  	if (_traceBooleanArray[y+_yOrigin][xp1+_xOrigin]) {
				  		coeffSum += coeffXp1Y;
				  	}
				  	else {
				  		chargeSum += chargeXp1Y;
				  	}
				  	double chargeXYp1 = charge * coeffXYp1;
				  	if (_traceBooleanArray[yp1+_yOrigin][x+_xOrigin]) {
				  		coeffSum += coeffXYp1;
				  	}
				  	else {
				  		chargeSum += chargeXYp1;
				  	}
				  	
				  	if (_traceBooleanArray[ym1+_yOrigin][x+_xOrigin]) {	
				    	chargeArrayTwo[ym1][x] += chargeXYm1;
				    	chargeArrayTwo[ym1][x] += chargeSum * coeffXYm1 / coeffSum;
				  	}
				  	if (_traceBooleanArray[y+_yOrigin][xm1+_xOrigin]) {
				    	chargeArrayTwo[y][xm1] += chargeXm1Y;	
				    	chargeArrayTwo[y][xm1] += chargeSum * coeffXm1Y / coeffSum;
				  	}
				  	if (_traceBooleanArray[y+_yOrigin][xp1+_xOrigin]) {
				    	chargeArrayTwo[y][xp1] += chargeXp1Y;	
				    	chargeArrayTwo[y][xp1] += chargeSum * coeffXp1Y / coeffSum;
				  	}
				  	if (_traceBooleanArray[yp1+_yOrigin][x+_xOrigin]) {
				    	chargeArrayTwo[yp1][x] += chargeXYp1;	
				    	chargeArrayTwo[yp1][x] += chargeSum * coeffXYp1 / coeffSum;
				  	}
				  	chargeArrayTwo[y][x] += charge * coeffXY;
				  	chargeArrayTwo[y][x] += chargeSum * coeffXY / coeffSum;
				}
			}
		}
	}
	
	
	/** 
	 * Returns the magnitude of the maximum electric field inside the charged fluid. This function
	 * also computes all the electric fields inside the charged fluid.
	 * @return	the magnitude of the maximum electric field.
	 */
	private double _getMaxElectricField()
	{
		double maxField = 0.0;		  	
		for (int y = _yStart; y < _yEnd; y++) {
			for (int x = _xStart; x < _xEnd; x++) {
				if (_traceBooleanArray[y+_yOrigin][x+_xOrigin]) {
					int xm1 = x - 1;
					int ym1 = y - 1;
					int xp1 = x + 1;
					int yp1 = y + 1;
					double electricPhiXM = _electricPotentialArray[y][xm1];
					double electricPhiXP = _electricPotentialArray[y][xp1];
					double electricPhiYM = _electricPotentialArray[ym1][x];
					double electricPhiYP = _electricPotentialArray[yp1][x];
					_electricFieldArrayInX[y][x] = (electricPhiXM - electricPhiXP);
					_electricFieldArrayInY[y][x] = (electricPhiYM - electricPhiYP);
					double magOfField = _electricFieldArrayInX[y][x] * _electricFieldArrayInX[y][x] +
								 		_electricFieldArrayInY[y][x] * _electricFieldArrayInY[y][x];
					if (magOfField > maxField) {
						maxField = magOfField;
					}
				}	
			}
		}
		return Math.sqrt(maxField);
	}
	
	
	/** 
	 * Sets the charges in the free charged fluid to <tt>_chargeArrayOne</tt>. 
	 */
	private void _setCharge2ArrayOne()
	{
		for (int y = _yStart; y < _yEnd; y++) {
			for (int x = _xStart; x < _xEnd; x++) {
				_chargeArrayOne[y][x] = 0.0;
			}
		}
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)it.next();
		  	Iterator itc = currentCF.linkedList.iterator();
		  	while (itc.hasNext()) {
		    	Charge currentCharge = (Charge)itc.next();
		    	Point chargePt = currentCharge.position;
		  		_chargeArrayOne[chargePt.y-_yOrigin][chargePt.x-_xOrigin] = currentCharge.charge;
		    }
		}
	}
	
	
	/** 
	 * Sets the charges in the given array to <tt>_electricPotentialArray</tt> for the
	 * electric potential computation via the FFT technique.
	 * @param	chargeArray		the specified charge array to be copied.
	 */
	private void _setElectricPotentialArray(double[][] chargeArray)
	{
		for (int y = 0; y < _arrayHeight; y++) {
			for (int x = 0; x < _arrayWidth; x++) {
				_electricPotentialArray[y][x] = chargeArray[y][x];
			}
		}
	}

	
	/** 
	 * Normalizes the charges of the free charged fluid for the purpose of merging multiple
	 * charged fluids in different <code>Poisson</code>. It can also control the electric
	 * potential of the free charged fluid to vary around the specified value -
	 * <tt>_normalizedPotential</tt>.
	 */
	public void normalizeChargedFluid()
	{
		for (int index = 0; index < linkedList.size(); index++) {
		  	ChargedFluid currentCF = (ChargedFluid)linkedList.get(index);
	  		double meanPotential = _getMeanPotential(index);
	  		double coeff = _normalizedPotential / meanPotential;
	  		double totalCharge = 0.0;
	      	Iterator it = currentCF.linkedList.iterator();
	      	while (it.hasNext()) {
	        	Charge currentCharge = (Charge)it.next();
	        	double newCharge = coeff * currentCharge.charge;
	        	totalCharge += newCharge;
	  			currentCharge.charge = newCharge;
	  		}
	  		currentCF.totalCharge = totalCharge;
		}
	}
	
	
	/** 
	 * Returns the mean electric potential in the specified charged fluid.
	 * @param	index	the index of the specified charged fluid.
	 * @return	the mean electric potential.
	 */
	private double _getMeanPotential(int index)
	{
		double meanPotential = 0.0;
		ChargedFluid currentCF = (ChargedFluid)linkedList.get(index);
      	Iterator it = currentCF.linkedList.iterator();
      	while (it.hasNext()) {
        	Charge currentCharge = (Charge)it.next();
        	double currentPotential = currentCharge.electricPotential;
        	meanPotential += currentPotential;
		}
		meanPotential /= currentCF.linkedList.size();
		return meanPotential;
	}
	
	
	/** 
	 * Constructs the 1-pixel-wide front based upon the boundary detection technique and
	 * the direction of the electric and effective fields.
	 */
	public void computeChargedFluidFront()
	{
		_chargeArray = _chargeArrayOne;  	
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
			double meanField = 0.0;
			int number = 0;
		  	ChargedFluid currentCF = (ChargedFluid)it.next();
	      	currentCF.linkedList.clear();
			for (int y = _yStart; y < _yEnd; y++) {
				for (int x = _xStart; x < _xEnd; x++) {
					if (_indexArray[y][x] >= 0) {
						int xMatrix = x + _xOrigin;
						int yMatrix = y + _yOrigin;
						Point chargePt = new Point(xMatrix, yMatrix);
						if (_is1PixelWideFront(chargePt)) {
							Charge newCharge = new Charge(_chargeArray[y][x], chargePt);
							double electricFieldInX = _electricFieldArrayInX[y][x];
							double electricFieldInY = _electricFieldArrayInY[y][x];
							newCharge.electricFieldInX = electricFieldInX;
							newCharge.electricFieldInY = electricFieldInY;
							double fieldMag = electricFieldInX * electricFieldInX
											+ electricFieldInY * electricFieldInY;
							fieldMag = Math.sqrt(fieldMag);
							newCharge.electricMagnitude = fieldMag;
				      		newCharge.electricPotential = _electricPotentialArray[y][x];
				      		if (!_isElectricDirectionInward(electricFieldInX, electricFieldInY, chargePt)) {
				      			meanField += fieldMag;
				      			number++;
				      		}
				      		else {
				      			newCharge.isElectricDirectionInverse = true;
				      		}
				      		currentCF.linkedList.add(newCharge);	
						}
					}
				}
			}
			meanField /= number;
	      	Iterator itc = currentCF.linkedList.iterator();
	      	while (itc.hasNext()) {
	        	Charge freeCharge = (Charge)itc.next();
	        	double eleMag = freeCharge.electricMagnitude;
	        	if (meanField > eleMag) {
		        	double meanFieldInX = meanField * freeCharge.electricFieldInX / eleMag;
		        	double meanFieldInY = meanField * freeCharge.electricFieldInY / eleMag;
		        	freeCharge.electricFieldInX = meanFieldInX;
		        	freeCharge.electricFieldInY = meanFieldInY;
	        	}
	        	if (freeCharge.isElectricDirectionInverse) {
	        		freeCharge.electricFieldInX = -freeCharge.electricFieldInX;
	        		freeCharge.electricFieldInY = -freeCharge.electricFieldInY;
	        	}
	        }
		}
		
		Iterator it2 = linkedList.iterator();
		while (it2.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)it2.next();
		  	Iterator itc2 = currentCF.linkedList.iterator();
	      	while (itc2.hasNext()) {
	        	Charge freeCharge = (Charge)itc2.next();
	        	Point chargePt = freeCharge.position;
	        	int xMatrix = chargePt.x;
	        	int yMatrix = chargePt.y;       	
	          	double imageFieldInX, imageFieldInY;
	          	if (_isGradientField) {
	          		imageFieldInX = _beta * _normalizedPotential * _gradientArrayInX[yMatrix][xMatrix];
	          		imageFieldInY = _beta * _normalizedPotential * _gradientArrayInY[yMatrix][xMatrix];
	          	}
	          	else {
		          	int xm1 = xMatrix - 1;
		          	int ym1 = yMatrix - 1;
		          	int xp1 = xMatrix + 1;
		          	int yp1 = yMatrix + 1;
		          	double imagePhiXM = _imageGradientArray[yMatrix][xm1];
		          	double imagePhiXP = _imageGradientArray[yMatrix][xp1];
		          	double imagePhiYM = _imageGradientArray[ym1][xMatrix];
		          	double imagePhiYP = _imageGradientArray[yp1][xMatrix]; 	
		          	imageFieldInX = _beta * _normalizedPotential * (imagePhiXP - imagePhiXM);
		          	imageFieldInY = _beta * _normalizedPotential * (imagePhiYP - imagePhiYM);
	          	}
	          	double effectiveFieldInX = freeCharge.electricFieldInX + imageFieldInX;
	          	double effectiveFieldInY = freeCharge.electricFieldInY + imageFieldInY;   	
	          	if (freeCharge.isElectricDirectionInverse) {
	          		effectiveFieldInX = freeCharge.electricFieldInX;
	          		effectiveFieldInY = freeCharge.electricFieldInY;
	          	}
	      		freeCharge.effectiveFieldInX = effectiveFieldInX;
	      		freeCharge.effectiveFieldInY = effectiveFieldInY;
	        }
		}
	}
	
	
	/** 
	 * Determines whether the specified point at the 2-pixel-wide interface is at 
	 * the 1-pixel-wide front, which is the boundary of the contour.
	 * @param	pt	the specified position to be tested.
	 * @return	<code>true</code> if the given point is at the boundary.
	 */
	private boolean _is1PixelWideFront(Point pt)
	{
		if (_traceBooleanArray[pt.y+1][pt.x-1] == false) {
			return true;
		}
		else if (_traceBooleanArray[pt.y+1][pt.x] == false) {
			return true;
		}
		else if (_traceBooleanArray[pt.y+1][pt.x+1] == false) {
			return true;
		}
		else if (_traceBooleanArray[pt.y][pt.x-1] == false) {
			return true;
		}
		else if (_traceBooleanArray[pt.y][pt.x+1] == false) {
			return true;
		}
		else if (_traceBooleanArray[pt.y-1][pt.x-1] == false) {
			return true;
		}
		else if (_traceBooleanArray[pt.y-1][pt.x] == false) {
			return true;
		}
		else if (_traceBooleanArray[pt.y-1][pt.x+1] == false) {
			return true;
		}
		else {
			return false;	
		}
	}
	
	
	/** 
	 * Detects whether the true number of the trace boolean array is reaching a constant.
	 * If the current number and the previous one are the same, set the free charged fluid
	 * frozen true.
	 */
	public void detectChargedFluidFrozen()
	{
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
			ChargedFluid currentCF = (ChargedFluid)it.next();
			if (!currentCF.isFrozen) {
			  	int minX = currentCF.minXAxis;
			  	int maxX = currentCF.maxXAxis;
			  	int minY = currentCF.minYAxis;
			  	int maxY = currentCF.maxYAxis;
				int nTraceCharge = 0;
				for (int y = minY - 1; y <= maxY + 1; y++) {
					for (int x = minX - 1; x <= maxX + 1; x++) {
						if (_traceBooleanArray[y][x] == true) {
							nTraceCharge++;
						}
					}
				}
				currentCF.previousNTraceCharge = currentCF.currentNTraceCharge;
				currentCF.currentNTraceCharge = nTraceCharge;
				if (currentCF.previousNTraceCharge == currentCF.currentNTraceCharge) {
					currentCF.isFrozen = true;
				}
			}
		}
	}
	
	
	/** 
	 * Determines whether the <code>Poisson</code> is frozen or not.
	 * @return	<code>true</code> if the free charged fluid is frozen.
	 */
  	public boolean isPoissonFrozen()
  	{
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)it.next();
		  	if (!currentCF.isFrozen) {
		  		return false;
		  	}
		}
		return true;
  	}
	
	
	/** 
	 * Trace the charges in the free charged fluid to create the contour in position order.
	 * A standard contour tracing algorithm is used to connect the charges at the outer
	 * boundary of the free charged fluid.
	 */
	public void traceChargedFluidFront()
	{
		ChargedFluid freeCF = (ChargedFluid)linkedList.getFirst();
      	Iterator itc = freeCF.linkedList.iterator();
      	while (itc.hasNext()) {
        	Charge freeCharge = (Charge)itc.next();
        	Point pt = freeCharge.position;
        	int xPoisson = pt.x - _xOrigin;
        	int yPoisson = pt.y - _yOrigin;
        	_freeChargeArray[yPoisson][xPoisson] = true;
        	_chargeArrayOne[yPoisson][xPoisson] = freeCharge.charge;
        	_electricFieldArrayInX[yPoisson][xPoisson] = freeCharge.effectiveFieldInX;
        	_electricFieldArrayInY[yPoisson][xPoisson] = freeCharge.effectiveFieldInY;
        	_electricPotentialArray[yPoisson][xPoisson] = freeCharge.electricPotential;
	    }
		_chargeArray = _chargeArrayOne;
		_startPosition = new Point(_xEnd, _yEnd);
		boolean hasStartPt = _contourStartPointDetection();
      	while (hasStartPt) {
      		boolean isContourClosed = false;
	      	_newChargedFluidLink = new LinkedList();
			_addNewCharge(_startPosition);
      		isContourClosed = _contourTracing(_startPosition, 1);
      		if (isContourClosed) {
      			_setBooleanContourArray();
      			linkedList.add(new ChargedFluid(_newChargedFluidLink));
      		}
      		else {
      			_setBooleanContourArray();
      			System.out.println("*** Contour tracing not closed ***");
      		}
      		hasStartPt = _contourStartPointDetection();
		}
		linkedList.removeFirst();
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
			ChargedFluid currentCF = (ChargedFluid)it.next();
			currentCF.computePolygon();
		}
	}
	
	
	/** 
	 * Returns the start point type for the contour tracing based on the search of the
	 * farthest point for each contour in the free charged fluid.
	 * @return	<code>true</code> if the start position can be found.
	 */
	private boolean _contourStartPointDetection()
	{
		for (int x = _startPosition.x; x >= _xStart; x--) {
			for (int y = _yEnd; y >= _yStart; y--) {
				if (_freeChargeArray[y][x]) {
					_startPosition = new Point(x,y);
					return true;
				}
			}
		}
		return false;
	}
	
	
	/** 
	 * Determines whether the traced contour is closed based on the given start point and
	 * the search direction type.
	 * @param	startPt		the start point to be tested.
	 * @param	direction	the searching direction according to the start point.
	 *						0: positive x-axis (->)
	 *						1: positive y-axis (/|\)
	 *						2: negative x-axis (<-)
	 *						3: negative y-axis (\|/)
	 * @return	<code>true</code> if the contour is closed.
	 */
	private boolean _contourTracing(Point startPt, int direction)
	{
		if (_newChargedFluidLink.size() >= 2) {
			Charge firstCharge = (Charge)_newChargedFluidLink.getFirst();
			Charge lastCharge = (Charge)_newChargedFluidLink.getLast();
			if (firstCharge.position.equals(lastCharge.position)) {
				_newChargedFluidLink.removeLast();
				return true;
			}
		}
		if (direction == 4) {
			direction = 0;
		}
		if (direction == -1) {
			direction = 3;
		}
		Point firstPt = new Point();
		Point secondPt = new Point();
		Point thirdPt = new Point();
		Point fourthPt = new Point();
		Point fifthPt = new Point();
		switch (direction) {
		  	case 0:
		  	  firstPt.setLocation(startPt.x+1,startPt.y-1);
		  	  secondPt.setLocation(startPt.x+1,startPt.y);
		  	  thirdPt.setLocation(startPt.x+1,startPt.y+1);
		  	  fourthPt.setLocation(startPt.x,startPt.y+1);
		  	  fifthPt.setLocation(startPt.x-1,startPt.y+1);
		  	  break;
		  	case 1:
		  	  firstPt.setLocation(startPt.x+1,startPt.y+1);
		  	  secondPt.setLocation(startPt.x,startPt.y+1);
		  	  thirdPt.setLocation(startPt.x-1,startPt.y+1);
		  	  fourthPt.setLocation(startPt.x-1,startPt.y);
		  	  fifthPt.setLocation(startPt.x-1,startPt.y-1);
		  	  break;
		  	case 2:
		  	  firstPt.setLocation(startPt.x-1,startPt.y+1);
		  	  secondPt.setLocation(startPt.x-1,startPt.y);
		  	  thirdPt.setLocation(startPt.x-1,startPt.y-1);
		  	  fourthPt.setLocation(startPt.x,startPt.y-1);
		  	  fifthPt.setLocation(startPt.x+1,startPt.y-1);
		  	  break;
		  	case 3:
		  	  firstPt.setLocation(startPt.x-1,startPt.y-1);
		  	  secondPt.setLocation(startPt.x,startPt.y-1);
		  	  thirdPt.setLocation(startPt.x+1,startPt.y-1);
		  	  fourthPt.setLocation(startPt.x+1,startPt.y);
		  	  fifthPt.setLocation(startPt.x+1,startPt.y+1);
		      break;
		    default:
		      System.out.println(" Contour tracing error");
		      return false;
		}
		if (_freeChargeArray[firstPt.y][firstPt.x]) {
			if (_freeChargeArray[secondPt.y][secondPt.x]) {
				_addNewCharge(secondPt);
			}
			_addNewCharge(firstPt);
			return _contourTracing(firstPt,direction-1);
		}
		else if (_freeChargeArray[secondPt.y][secondPt.x]) {
			_addNewCharge(secondPt);
			return _contourTracing(secondPt,direction);
		}
		else if (_freeChargeArray[thirdPt.y][thirdPt.x]) {
			if (_freeChargeArray[fourthPt.y][fourthPt.x]) {
				_addNewCharge(fourthPt);
			}
			_addNewCharge(thirdPt);
			return _contourTracing(thirdPt,direction);
		}
		else if (_freeChargeArray[fourthPt.y][fourthPt.x] || 
					_freeChargeArray[fifthPt.y][fifthPt.x]) {
			return _contourTracing(startPt,direction+1);
		}
		else {
			return false;
		}
	}
	
	
	/**
	 * Adds a new charge to the <tt>_newChargedFluidLink</tt> based upon the given point.
	 * @param	pt	the new charge position.
	 */
	private void _addNewCharge(Point pt)
	{
		Charge newCharge = new Charge(_chargeArray[pt.y][pt.x], new Point(pt.x+_xOrigin, pt.y+_yOrigin));
  		newCharge.effectiveFieldInX = (_electricFieldArrayInX[pt.y][pt.x]);
  		newCharge.effectiveFieldInY = (_electricFieldArrayInY[pt.y][pt.x]);
	    newCharge.electricPotential = (_electricPotentialArray[pt.y][pt.x]);
		_newChargedFluidLink.add(newCharge);
	}
	
	
	/** 
	 * Sets false to the charge position in the <tt>_freeChargeArray</tt>. The charges are in
	 * the traced charged fluid, <tt>_newChargedFluidLink</tt>, that is created right after
	 * the contour tracing procedure. This function is used to disable the positions of
	 * the traced charges in the free charged fluid for multiple charged fluids.
	 */
	private void _setBooleanContourArray()
	{
		Iterator it = _newChargedFluidLink.iterator();
		while (it.hasNext()) {
			Charge currentCharge = (Charge)it.next();
		    Point chargePt = currentCharge.position;
		    _freeChargeArray[chargePt.y-_yOrigin][chargePt.x-_xOrigin] = false;
		}
	}
	
	
	/** 
	 * Sets uniform charge to the charges of the free charged fluid. 
	 */	
	private void _setUniformCharge2ChargedFluid()
	{
		Iterator it = linkedList.iterator();
		while (it.hasNext()) {
		  	ChargedFluid currentCF = (ChargedFluid)it.next();	
	  		int numberOfCharge = currentCF.linkedList.size();
	  		double charge = currentCF.totalCharge / (double)numberOfCharge;
	      	Iterator itc = currentCF.linkedList.iterator();
	      	while (itc.hasNext()) {
	        	Charge currentCharge = (Charge)itc.next();
	        	currentCharge.charge = charge;
	        }
		}
	}
}